# safe_import_utils.py
import sys, importlib, os, inspect
from pathlib import Path

# --- tiny utils ---
def find_top_pkg_dir(p: Path) -> Path | None:
    p = p.resolve()
    if p.is_file():
        p = p.parent
    top = None
    while (p / "__init__.py").exists():
        top = p
        if p.parent == p:
            break
        p = p.parent
    return top

def derive_package_for_file(file: str) -> tuple[str, Path]:
    """Return (pkg_name, sysroot) for the module file."""
    here = Path(file).resolve()
    top_pkg_dir = find_top_pkg_dir(here)
    if not top_pkg_dir:
        raise RuntimeError(f"No package context above {here}. Add __init__.py up the tree.")
    sysroot = top_pkg_dir.parent
    if str(sysroot) not in sys.path:
        sys.path.insert(0, str(sysroot))
    parts = here.with_suffix("").relative_to(sysroot).parts
    pkg_name = ".".join(parts[:-1])  # package of the module (drop the filename)
    return pkg_name, sysroot

def ensure_caller_package(caller_file: str, caller_globals: dict | None = None) -> str:
    """
    Ensure sysroot is on sys.path and return the caller's dotted package.
    Optionally set caller_globals['__package__'] when running as a script.
    """
    pkg, _ = derive_package_for_file(caller_file)
    if caller_globals and caller_globals.get("__name__") == "__main__" and not caller_globals.get("__package__"):
        caller_globals["__package__"] = pkg
    return pkg

# --- public API ---
def safe_import(
    name: str,
    *,
    member: str | None = None,
    package: str | None = None,
    file: str | None = None,
    caller_globals: dict | None = None,
):
    """
    Import `name` (relative or absolute).
    - If `name` is relative and `package` is missing, derive it from `file` (or caller).
    - If running the caller as a script, we can set its __package__ when caller_globals is provided.
    """
    if file is None:
        # best-effort: use the immediate caller's file
        frame = inspect.currentframe()
        assert frame is not None
        outer = frame.f_back
        caller_file = (outer.f_globals.get("__file__") if outer else None) or __file__
    else:
        caller_file = file

    if name.startswith(".") and not package:
        package = ensure_caller_package(caller_file, caller_globals=caller_globals)

    mod = importlib.import_module(name, package=package)
    return getattr(mod, member) if member else mod



from ...read_write_utils import read_from_file,write_to_file
from ...string_clean import eatAll,eatElse,clean_line
from ...class_utils import get_caller_path
from ...list_utils import make_list
import os
import_tag = 'import '
from_tag = 'from '
def get_text_or_read(text=None,file_path=None):
    text = text or ''
    imports_js = {}
    if not text and file_path and os.path.isfile(file_path):
        text=read_from_file(file_path)
    return text
def is_line_import(line):
    if line and (line.startswith(from_tag) or line.startswith(import_tag)):
        return True
    return False
def is_line_group_import(line):
    if line and (line.startswith(from_tag) and import_tag in line):
        return True
    return False
def get_import_pkg(line):
    if is_line_group_import(line):
        return clean_line(line.split(from_tag)[1].split(import_tag)[0])
def get_imports_from_import_pkg(line):
    if is_line_group_import(line):
        return get_cleaned_import_list(line,commaClean=True)

def add_imports_to_import_pkg_js(import_pkg,imports,import_pkg_js=None):
    import_pkg_js = import_pkg_js or {}
    imports = clean_imports(imports)
    if import_pkg not in import_pkg_js:
        i = len(import_pkg_js["nulines"])
        import_pkg_js[import_pkg]={"imports":imports,"line":i}
        import_line = f"from {import_pkg} import "
        if import_pkg == "import":
            import_line = import_tag
        import_pkg_js["nulines"].append(import_line)
    else:
        import_pkg_js[import_pkg]["imports"]+=imports
    return import_pkg_js
def update_import_pkg_js(line,import_pkg_js=None):
    import_pkg_js = import_pkg_js or {}
    if is_line_group_import(line):
        import_pkg = get_import_pkg(line)
        imports = get_imports_from_import_pkg(line)
        import_pkg_js = add_imports_to_import_pkg_js(import_pkg,imports,import_pkg_js=import_pkg_js)
    else:
        if len(import_pkg_js["nulines"]) >0 and line == '' and is_line_import(import_pkg_js["nulines"][-1]):
            pass
        else:
            import_pkg_js["nulines"].append(line)
    return import_pkg_js
def is_from_line_group(line):
    if line and line.startswith(from_tag) and import_tag in line and '(' in line:
        import_spl = line.split(import_tag)[-1]
        import_spl_clean = clean_line(line)
        if not import_spl_clean.endswith(')'):
            return True
    return False
def clean_imports(imports,commaClean=True):
    chars=["*"]
    if not commaClean:
        chars.append(',')
    if isinstance(imports,str):
        imports = imports.split(',')
    return [eatElse(imp,chars=chars) for imp in imports if imp]
def get_cleaned_import_list(line,commaClean=True):
    cleaned_import_list=[]
    if import_tag in line:
        imports = line.split(import_tag)[1]
        cleaned_import_list+=clean_imports(imports,commaClean=commaClean)
    return cleaned_import_list
def get_all_imports(text=None,file_path=None,import_pkg_js=None):
    text = get_text_or_read(text=text,file_path=file_path)
    lines = text.split('\n')
    cleaned_import_list=[]
    nu_lines = []
    is_from_group = False
    import_pkg_js = import_pkg_js or {}
    if "nulines" not in import_pkg_js:
        import_pkg_js["nulines"]=[]
    if "file_path" not in import_pkg_js:
        import_pkg_js["file_path"]=file_path
    if "all_data" not in import_pkg_js:
        import_pkg_js["all_data"]=[]
    if file_path and file_path != import_pkg_js["file_path"]:
        found=False
        nu_data = {"file_path":import_pkg_js["file_path"],"nulines":import_pkg_js["nulines"]}
        for i,data in enumerate(import_pkg_js["all_data"]):
            if data.get('file_path') == import_pkg_js["file_path"]:
                import_pkg_js["all_data"][i] = nu_data
                found = True
                break
        if found == False:
            import_pkg_js["all_data"].append(nu_data)
        import_pkg_js["nulines"]=[]
        import_pkg_js["file_path"]=file_path

    for line in lines:
        if line.startswith(import_tag) and ' from ' not in line:
            cleaned_import_list = get_cleaned_import_list(line)
            import_pkg_js = add_imports_to_import_pkg_js("import",cleaned_import_list,import_pkg_js=import_pkg_js)
        else:
            if is_from_group:
                import_pkg=is_from_group
                line = clean_line(line)
                if line.endswith(')'):
                   is_from_group=False
                   line=line[:-1]
                imports_from_import_pkg = clean_imports(line)
                import_pkg_js = add_imports_to_import_pkg_js(import_pkg,imports_from_import_pkg,import_pkg_js=import_pkg_js)
                
            else:
                import_pkg_js=update_import_pkg_js(line,import_pkg_js=import_pkg_js)
            if is_from_line_group(line) and is_from_group == False:
                is_from_group=get_import_pkg(line)
    return import_pkg_js
def clean_all_imports(text=None,file_path=None,import_pkg_js=None):
    if not import_pkg_js:
        import_pkg_js = get_all_imports(text=text,file_path=file_path)
    nu_lines = import_pkg_js["nulines"]
    for pkg,values in import_pkg_js.items():
        comments = []
        if pkg not in ["nulines","file_path","all_data"]: 
            line = values.get('line')
            imports = values.get('imports')
            for i,imp in enumerate(imports):
                if '#' in imp:
                    imp_spl = imp.split('#')
                    comments.append(imp_spl[-1])
                    imports[i] = clean_line(imp_spl[0])
            imports = list(set(imports))   
            if '*' in imports:
                imports="*"
            else:
                imports=','.join(imports)
                if comments:
                    comments=','.join(comments)
                    imports+=f" #{comments}"
            import_pkg_js[pkg]["imports"]=imports
            nu_lines[line] += imports
    import_pkg_js["nulines"]=nu_lines
    return import_pkg_js

def get_all_real_imps(file):
    contents = read_from_file(file)
    lines = contents.split('\n')
    for line in lines:
        if line.startswith('from '):
            from_line = line.split('from ')[-1]
            dot_fro = ""
            dirname = file
            for char in from_line:
                if  char != '.':
                    line = f"from {dot_fro}{eatAll(from_line,'.')}"
                    if line in all_imps:
                        line = ""
                    break
                if dot_fro == "":
                    dot_fro = ""
                dirname = os.path.dirname(dirname)
                dirbase = os.path.basename(dirname)
                dot_fro = f"{dirbase}.{dot_fro}"
        if line:
            all_imps.append(line)

    return '\n'.join(all_imps)
def save_cleaned_imports(text=None,file_path=None,write=False,import_pkg_js=None):
    import_pkg_js=get_all_imports(text=text,file_path=file_path,import_pkg_js=import_pkg_js)
    import_pkg_js = clean_all_imports(text=text,file_path=file_path,import_pkg_js=import_pkg_js)
    contents = '\n'.join(import_pkg_js["nulines"])
    if file_path and write:
        write_to_file(contents=contents,file_path=file_path)
    return contents



from .list_utils import make_list
def get_from_kwargs(*args,**kwargs):
    del_kwarg = kwargs.get('del_kwargs',False)
    values = {}
    for key in args:
        if key:
            key = str(key)
            if key in kwargs:
                values[key] = kwargs.get(key)
                if del_kwarg:
                    del kwargs[key]
    return values,kwargs

def replace_it(string,item,rep):
    if item in string:
        string = string.replace(item,rep)
    return string
def while_replace(string,item,rep):
    while True:
        string = replace_it(string,item,rep)
        if item not in string or item in rep:
            return string
def for_replace(string,item,replace):
    replace = make_list(replace)
    for rep in replace:
        string = while_replace(string,item,rep)
    return string
def replace_all(string,*args,**kwargs):
    for items in args:
        if items and isinstance(items,list):
            item = items[0]
            replace = items[1:] if len(items)>1 else items[-1]
            string = for_replace(string,item,replace)
    values,kwargs = get_from_kwargs('item','replace',**kwargs)
    if values:
        string = for_replace(string,**values)
    for item,replace in kwargs.items():
        string = for_replace(string,item,rep)
    return string
def get_lines(string,strip=True):
    lines = string.split('\n')
    if strip:
        lines = [line for line in lines if line]
    return lines
def get_alpha():
    return 'abcdefghijklmnopqrstuvwxyz'
def is_alpha(char,case_sensative=False):
    alphas = get_alpha()
    if not case_sensative:
        alphas+=alphas.upper()
    return char in alphas



# ============================================================
# abstract_utilities/imports/imports.py
# Global imports hub — everything imported here will be
# automatically available to any module that does:
#     from ..imports import *
# ============================================================


import os
import sys, importlib,os
import sys, importlib, os, inspect
from pathlib import Path
import os,sys



from typing import *
import re

from typing import *
from types import MethodType
import os,re, sys, importlib, inspect, os, importlib.util, hashlib
import os,tempfile,shutil,logging,ezodf,fnmatch,pytesseract,pdfplumber
import pandas as pd
import geopandas as gpd
from datetime import datetime

from typing import *
from werkzeug.utils import secure_filename
from werkzeug.datastructures import FileStorage
from pdf2image import convert_from_path   # only used for OCR fallback
# ---- Core standard library modules -------------------------
import os, sys, re, shlex, glob, platform, textwrap, subprocess, inspect, json, time
import tempfile, shutil, logging, pathlib, fnmatch, importlib, importlib.util, types

from datetime import datetime
from types import ModuleType

# ---- Dataclasses and typing --------------------------------
from dataclasses import dataclass, field
from typing import (
    Any, Optional, List, Dict, Set, Tuple,
    Iterable, Callable, Literal, Union, TypeVar
)

# ---- Common 3rd-party dependencies --------------------------
import pandas as pd
import geopandas as gpd
import pytesseract
import pdfplumber
import PyPDF2
import ezodf
from pdf2image import convert_from_path
from werkzeug.utils import secure_filename
from werkzeug.datastructures import FileStorage

# ---- Helpers ------------------------------------------------
import textwrap as tw
from pprint import pprint

# ============================================================
# AUTO-EXPORT ALL NON-PRIVATE NAMES
# ============================================================
__all__ = [name for name in globals() if not name.startswith("_")]




from .file_filters import *
from .file_utils import *
from .filter_params import *
from .map_utils import *
from .pdf_utils import *
from .file_reader import *
from .find_collect import *
from .type_checks import *



from .imports import *
from .utils import run_local_cmd, run_ssh_cmd,run_any_cmd,run_cmd

class PathBackend(Protocol):
    def join(self, *parts: str) -> str: ...
    def isfile(self, path: str) -> bool: ...
    def isdir(self, path: str) -> bool: ...
    def glob_recursive(self, base: str, **opts) -> List[str]: ...
    def listdir(self, base: str) -> List[str]: ...

class LocalFS:
    def join(self, *parts: str) -> str:
        return os.path.join(*parts)
    def isfile(self, path: str) -> bool:
        return os.path.isfile(path)
    def isdir(self, path: str) -> bool:
        return os.path.isdir(path)
    def glob_recursive(self, base: str, **opts) -> List[str]:
        """
        opts:
          - maxdepth: int | None
          - mindepth: int (default 1)
          - follow_symlinks: bool
          - include_dirs: bool
          - include_files: bool
          - exclude_hidden: bool
        """
        maxdepth = opts.get("maxdepth")
        mindepth = opts.get("mindepth", 1)
        follow   = opts.get("follow_symlinks", False)
        want_d   = opts.get("include_dirs", True)
        want_f   = opts.get("include_files", True)
        hide     = opts.get("exclude_hidden", False)

        results: List[str] = []
        base_depth = os.path.normpath(base).count(os.sep)

        for root, dirs, files in os.walk(base, followlinks=follow):
            depth = os.path.normpath(root).count(os.sep) - base_depth
            if maxdepth is not None and depth > maxdepth:
                dirs[:] = []
                continue
            if want_d and depth >= mindepth:
                for d in dirs:
                    if hide and d.startswith("."): continue
                    results.append(os.path.join(root, d))
            if want_f and depth >= mindepth:
                for f in files:
                    if hide and f.startswith("."): continue
                    results.append(os.path.join(root, f))
        return results

    def listdir(self, base: str) -> List[str]:
        try:
            return [os.path.join(base, name) for name in os.listdir(base)]
        except Exception:
            return []

class SSHFS:
    """Remote POSIX backend via your run_remote_cmd."""
    def __init__(self, user_at_host: str):
        self.user_at_host = user_at_host

    def join(self, *parts: str) -> str:
        return posixpath.join(*parts)

    def isfile(self, path: str) -> bool:
        cmd = f"test -f {shlex.quote(path)} && echo __OK__ || true"
        out = run_remote_cmd(self.user_at_host, cmd)
        return "__OK__" in (out or "")

    def isdir(self, path: str) -> bool:
        cmd = f"test -d {shlex.quote(path)} && echo __OK__ || true"
        out = run_remote_cmd(self.user_at_host, cmd)
        return "__OK__" in (out or "")

    def glob_recursive(self, base: str, **opts) -> List[str]:
        maxdepth = opts.get("maxdepth")
        mindepth = opts.get("mindepth", 1)
        follow   = opts.get("follow_symlinks", False)
        want_d   = opts.get("include_dirs", True)
        want_f   = opts.get("include_files", True)
        hide     = opts.get("exclude_hidden", False)

        parts = []
        if follow:
            parts.append("-L")
        parts += ["find", shlex.quote(base)]
        if mindepth is not None:
            parts += ["-mindepth", str(mindepth)]
        if maxdepth is not None:
            parts += ["-maxdepth", str(maxdepth)]

        type_filters = []
        if want_d and not want_f:
            type_filters = ["-type", "d"]
        elif want_f and not want_d:
            type_filters = ["-type", "f"]

        hidden_filter = []
        if hide:
            hidden_filter = ["!", "-regex", r".*/\..*"]

        cmd = " ".join(parts + type_filters + hidden_filter + ["-printf", r"'%p\n'"]) + " 2>/dev/null"
        out = run_remote_cmd(self.user_at_host, cmd)
        return [line.strip().strip("'") for line in (out or "").splitlines() if line.strip()]

    def listdir(self, base: str) -> List[str]:
        cmd = f"find {shlex.quote(base)} -maxdepth 1 -mindepth 1 -printf '%p\\n' 2>/dev/null"
        out = run_remote_cmd(self.user_at_host, cmd)
        return [line.strip() for line in (out or "").splitlines() if line.strip()]

# ---- auto-detect "user@host:/abs/path" ----
REMOTE_RE = re.compile(r"^(?P<host>[^:\s]+@[^:\s]+):(?P<path>/.*)$")

def normalize_items(paths: Iterable[str],user_at_host=None,**kwargs) -> List[tuple[PathBackend, str]]:
    pairs: List[tuple[PathBackend, str]] = []
    host = user_at_host or kwargs.get('host')
    
    for item in paths:
        if not item: continue
        m = REMOTE_RE.match(item)
        if m:
            pairs.append((SSHFS(m.group("host") or user_at_host), m.group("path")))
        else:
            pairs.append((LocalFS(), item))
    return pairs



"""
read_write_utils.py
-------------------
Unified read/write utility for safe file operations.
Supports:
- Writing content to a file
- Reading content from a file
- Creating and reading if missing
- Detecting file/content params via positional args or kwargs

Usage:
    from abstract_utilities.read_write_utils import *
"""

import os
import shlex
from .ssh_utils.utils import run_cmd,get_print_sudo_cmd,run_local_cmd,run_remote_cmd
from .file_utils.file_utils.type_checks import is_file,is_dir,get_user_pass_host_key,is_exists
from .abstract_classes import run_pruned_func
from .string_utils import get_from_kwargs
_FILE_PATH_KEYS = ['file', 'filepath', 'file_path', 'path', 'directory', 'f', 'dst', 'dest']
_CONTENTS_KEYS = ['cont', 'content', 'contents', 'data', 'datas', 'dat', 'src', 'source']

            
# --- Helper utilities --------------------------------------------------------
def string_in_keys(strings, kwargs):
    """Find a matching keyword in kwargs that contains any of the given substrings."""
    for key in kwargs:
        for s in strings:
            if s.lower() in key.lower():
                return key
    return None
def make_dirs(path, exist_ok=True, **kwargs):
    remote = get_user_pass_host_key(**kwargs)
    print(remote)
    if remote:
        kwargs['cmd'] = f"mkdir -p {path}"
        print(kwargs)
        resp = run_pruned_func(run_cmd, **kwargs)
        print(resp)
    else:
        os.makedirs(path, exist_ok=exist_ok)
    return path
def make_path(path, home_dir=None, file=None, **kwargs):
    if not path:
        return None

    basename = os.path.basename(path)
    parts = [p for p in path.split('/') if p]

    # Detect whether this is a file or a folder
    is_file = file if file is not None else ('.' in basename)
    pieces = parts[:-1] if is_file else parts
    print(pieces)
    full_dir = home_dir or '/'
    for piece in pieces:
        full_dir = os.path.join(full_dir, piece)
        make_dirs(full_dir, exist_ok=True, **kwargs)
        print(f"✅ full_dir == {full_dir}")
    if is_file:
        full_dir = os.path.join(full_dir, basename)

    print(f"✅ full_dir == {full_dir}")
    return full_dir
def get_rel_path(src,src_rel,dst,**kwargs):
    if src.startswith(src_rel):
        nu_src = src[len(src_rel):]
        nu_src= eatAll(nu_src,'/')
        directory= eatOuter(dst,'/')
        rel_path = os.path.join(dst,nu_src)
        return rel_path
def make_relative_path(src,src_rel,dst,**kwargs):
    print(f"src == {src}\nsrc_rel == {src_rel}\dst == {dst}")
    if src.startswith(src_rel):
        rel_path = get_rel_path(src,src_rel,dst)
        print(rel_path)
        path = make_path(rel_path,**kwargs)
        print(f"path == {path}")
        return path

def path_join(*args):
    path = None
    for i,arg in enumerate(args):
        if arg:
            if i == 0:
                path = arg
            else:
                path = os.path.join(path,arg)
    return path

def get_path(paths,**kwargs):
    """Return the first valid path among given paths."""
    for path in paths:
        if isinstance(path, str):
            if is_file(path,**kwargs):
                return path
            dirname = os.path.dirname(path)
            if is_exists(dirname,**kwargs):
                return path
    return None


def break_down_find_existing(path,**kwargs):
    """Return the first non-existent subpath within a path chain."""
    test_path = ''
    for part in path.split(os.sep):
        test_path = os.path.join(test_path, part)
        if not is_exists(test_path,**kwargs):
            return test_path if test_path else None
    return test_path


# --- Parameter parsing --------------------------------------------------------
def check_read_write_params(*args, **kwargs):
    """
    Determine file_path and contents from arguments.
    Returns a tuple: (file_path, contents)
    """
    file_key = string_in_keys(_FILE_PATH_KEYS, kwargs)
    content_key = string_in_keys(_CONTENTS_KEYS, kwargs)

    file_path = kwargs.get(file_key) if file_key else None
    contents = kwargs.get(content_key) if content_key else None

    # Handle positional args (fallback)
    if file_path is None and len(args) > 0:
        file_path = args[0]
    if contents is None and len(args) > 1:
        contents = args[1]

    if file_path is None:
        raise ValueError("Missing file_path argument.")
    return file_path, contents

def write_to_path(
        file_path: str,
        contents: str,
        *,
        user_at_host: str = None,
        cwd: str | None = None,
        password=None,
        key=None,
        env_path=None,
        **kwargs
    ) -> str:
    """
    Completely overwrite a file (locally or remotely).
    Supports sudo and password-based remote execution.
    """

    # sanitize for shell safety
    quoted_path = shlex.quote(file_path)
    quoted_data = shlex.quote(str(contents))

    # shell command that fully overwrites
    # (no append, replaces contents entirely)
    base_cmd = f'sudo sh -c "echo {quoted_data} > {quoted_path}"'
    input(base_cmd)
    # optional sudo password injection
    full_cmd = get_print_sudo_cmd(
        cmd=base_cmd,
        password=password,
        key=key,
        env_path=env_path
    )

    # local or remote dispatch
    if user_at_host:
        return run_remote_cmd(
            user_at_host=user_at_host,
            cmd=full_cmd,
            cwd=cwd,
            password=password,
            key=key,
            env_path=env_path,
            **kwargs
        )
    else:
        return run_local_cmd(
            cmd=full_cmd,
            cwd=cwd,
            password=password,
            key=key,
            env_path=env_path,
            **kwargs
        )
### --- Core functionality -------------------------------------------------------
##def write_to_file(*args, **kwargs):
##    """
##    Write contents to a file (create if missing).
##
##    Returns the file_path written.
##    """
##    file_path, contents = check_read_write_params(*args, **kwargs)
##    if contents is None:
##        raise ValueError("Missing contents to write.")
##
##    os.makedirs(os.path.dirname(file_path) or ".", exist_ok=True)
##    with open(file_path, "w", encoding="utf-8") as f:
##        f.write(str(contents))
##    return file_path
# --- Core functionality -------------------------------------------------------
def write_to_file(*args, **kwargs):
    """
    Write contents to a file (create if missing).

    Returns the file_path written.
    """
    file_path, contents = check_read_write_params(*args, **kwargs)
    values,kwargs = get_from_kwargs(['file_path','contents'],del_kwarg=True,**kwargs)
    dirname = os.path.dirname(file_path)
    
    if contents is None:
        raise ValueError("Missing contents to write.")
    user_at_host = kwargs.get("user_at_host")
    if get_user_pass_host_key(**kwargs):
        make_dirs(dirname, exist_ok=True,**kwargs)
        kwargs["cwd"] = kwargs.get('cwd') or os.path.dirname(file_path)
        # sanitize for shell safety
        quoted_path = shlex.quote(file_path)
        quoted_data = shlex.quote(str(contents))
        # shell command that fully overwrites
        # (no append, replaces contents entirely)
        kwargs["cmd"] = f'sh -c "echo {quoted_data} > {quoted_path}"'
        if not kwargs.get('password') and not kwargs.get('key'):
            kwargs["cmd"]=f'sudo {kwargs["cmd"]}'
        result = run_pruned_func(run_cmd,**kwargs)
        if 'file_path' in kwargs:
            del kwargs['file_path']
        if not is_file(file_path,**kwargs) or str(contents) != read_from_file(file_path,**kwargs):
            kwargs["cmd"]=f'sudo {kwargs["cmd"]}'
            result = run_pruned_func(run_cmd,**kwargs)
        return result

    make_dirs(dirname or ".", exist_ok=True)
    with open(file_path, "w", encoding="utf-8") as f:
        f.write(str(contents))
    return file_path


def read_from_file(file_path,**kwargs):
    if get_user_pass_host_key(**kwargs):
        kwargs["cwd"] = kwargs.get('cwd') or os.path.dirname(file_path)
        basename = os.path.basename(file_path)
        kwargs["cmd"] = f'cat {basename}'
        return run_pruned_func(run_cmd,**kwargs)
    """Read text content from a file."""
    with open(file_path, "r", encoding="utf-8") as f:
        return f.read()

def copy_dirs(dirs,dst,src_rel=None,**kwargs):
    for src in dirs:
        if rel_path:
            dst = make_relative_path(src,src_rel,dst,**kwargs)
        make_path(dst,**kwargs)
        
def copy_file(src,dst,rel_path=None,**kwargs):

    if rel_path:
        dst = make_relative_path(src,rel_path,dst,**kwargs)
        print(dst)
    if get_user_pass_host_key(**kwargs):
        contents=read_from_file(src,**kwargs)
        write_to_file(contents=contents,file_path=dst,**kwargs)
    else:
        shutil.copy(src,dst)
    print(dst)
    return dst
def copy_files(files,dst,rel_path=None,**kwargs):
    for file in files:
        copy_file(src=file,dst=dst,rel_path=rel_path,**kwargs)

def create_and_read_file(*args, **kwargs):
    """
    Create the file (if missing) and read contents from it.
    """
    file_path, contents = check_read_write_params(*args, **kwargs)
    if not os.path.isfile(file_path):
        write_to_file(file_path, contents or "")
    return read_from_file(file_path)


def is_file_extension(obj: str) -> bool:
    """Return True if obj looks like a filename with extension."""
    if not isinstance(obj, str):
        return False
    root, ext = os.path.splitext(obj)
    return bool(root and ext)


def delete_file(file_path: str):
    """Safely delete a file if it exists."""
    if os.path.isfile(file_path):
        os.remove(file_path)
        return True
    return False


def get_content_lines(*args, **kwargs):
    """Return a list of lines from string or file path."""
    file_path, contents = check_read_write_params(*args, **kwargs)
    if os.path.isfile(file_path):
        contents = read_from_file(filepath)

    if isinstance(contents, str):
        return contents.splitlines()
    elif isinstance(contents, list):
        return contents
    return []



from .imports import *
from ....string_clean import eatAll
from ....list_utils import make_list
from ....type_utils import get_media_exts, is_media_type, MIME_TYPES, is_str
from ....ssh_utils import *
from ....env_utils import *
from ....read_write_utils import *
from ....abstract_classes import SingletonMeta
from ....string_utils import get_from_kwargs
from ....abstract_classes import run_pruned_func
from ....class_utils import get_caller, get_caller_path, get_caller_dir


__all__ = [name for name in globals() if not name.startswith("_")]




from .filter_params import *
from .imports import *

##from abstract_utilities import make_list,get_media_exts, is_media_type
def get_allowed_predicate(allowed=None):
    if allowed != False:
        if allowed == True:
            allowed = None
        allowed = allowed or make_allowed_predicate()
    else:
        def allowed(*args):
            return True
        allowed = allowed
    return allowed
def get_globs(items,recursive: bool = True,allowed=None):
    glob_paths = []
    items = [item for item in make_list(items) if item]
    for item in items:
        pattern = os.path.join(item, "**/*")  # include all files recursively\n
        nuItems = glob.glob(pattern, recursive=recursive)
        if allowed:
            nuItems = [nuItem for nuItem in nuItems if nuItem and allowed(nuItem)]
        glob_paths += nuItems
    return glob_paths
def get_allowed_files(items,allowed=True):
    allowed = get_allowed_predicate(allowed=allowed)
    return [item for item in items if item and os.path.isfile(item) and allowed(item)]
def get_allowed_dirs(items,allowed=False):
    allowed = get_allowed_predicate(allowed=allowed)
    return [item for item in items if item and os.path.isdir(item) and allowed(item)]

def get_filtered_files(items,allowed=None,files = []):
    allowed = get_allowed_predicate(allowed=allowed)
    glob_paths = get_globs(items)
    return [glob_path for glob_path in glob_paths if glob_path and os.path.isfile(glob_path) and glob_path not in files and allowed(glob_path)]
def get_filtered_dirs(items,allowed=None,dirs = []):
    allowed = get_allowed_predicate(allowed=allowed)
    glob_paths = get_globs(items)
    return [glob_path for glob_path in glob_paths if glob_path and os.path.isdir(glob_path) and glob_path not in dirs and allowed(glob_path)]

def get_all_allowed_files(items,allowed=None):
    dirs = get_all_allowed_dirs(items)
    files = get_allowed_files(items)
    nu_files = []
    for directory in dirs:
        files += get_filtered_files(directory,allowed=allowed,files=files)
    return files
def get_all_allowed_dirs(items,allowed=None):
    allowed = get_allowed_predicate(allowed=allowed)
    dirs = get_allowed_dirs(items)
    nu_dirs=[]
    for directory in dirs:
        nu_dirs += get_filtered_dirs(directory,allowed=allowed,dirs=nu_dirs)
    return nu_dirs
def get_files_and_dirs(
    directory: str,
    cfg: Optional["ScanConfig"] = None,
    allowed_exts: Optional[Set[str]] = False,
    unallowed_exts: Optional[Set[str]] = False,
    exclude_types: Optional[Set[str]] = False,
    exclude_dirs: Optional[List[str]] = False,
    exclude_patterns: Optional[List[str]] = False,
    add = False,
    recursive: bool = True,
    include_files: bool = True,
    **kwargs
    ):
    cfg = cfg or define_defaults(
        allowed_exts=allowed_exts,
        unallowed_exts=unallowed_exts,
        exclude_types=exclude_types,
        exclude_dirs=exclude_dirs,
        exclude_patterns=exclude_patterns,
        add=add
        )
    allowed = make_allowed_predicate(cfg)
    items=[]
    files =[]
    if recursive:
        items = get_globs(directory,recursive=recursive,allowed=allowed)
    else:
        directories = make_list(directory)
        for directory in directories:
            items +=[os.path.join(directory,item) for item in os.listdir(directory)]
    dirs = get_allowed_dirs(items,allowed=allowed)
    if include_files:
        files = get_allowed_files(items,allowed=allowed)
    return dirs,files
def make_allowed_predicate(cfg: ScanConfig) -> Callable[[str], bool]:
    def allowed(path: str) -> bool:
        p = Path(path)
        name = p.name.lower()
        path_str = str(p).lower()
        # A) directories
        if cfg.exclude_dirs:
            for dpat in cfg.exclude_dirs:
                if dpat in path_str or fnmatch.fnmatch(name, dpat.lower()):
                    if p.is_dir() or dpat in path_str:
                        return False

        if cfg.exclude_patterns:
            # B) filename patterns
            for pat in cfg.exclude_patterns:
                if fnmatch.fnmatch(name, pat.lower()):
                    return False

        # C) extension gates
        if p.is_file():
            ext = p.suffix.lower()
            if (cfg.allowed_exts and ext not in cfg.allowed_exts) or (cfg.unallowed_exts and ext in cfg.unallowed_exts):
                return False
        return True
    return allowed
def collect_filepaths(
    directory: List[str],
    cfg: ScanConfig=None,
    allowed_exts: Optional[Set[str]] = False,
    unallowed_exts: Optional[Set[str]] = False,
    exclude_types: Optional[Set[str]] = False,
    exclude_dirs: Optional[List[str]] = False,
    exclude_patterns: Optional[List[str]] = False,
    add=False,
    allowed: Optional[Callable[[str], bool]] = None,
    **kwargs
    ) -> List[str]:
    cfg = cfg or define_defaults(
                                allowed_exts=allowed_exts,
                                unallowed_exts=unallowed_exts,
                                exclude_types=exclude_types,
                                exclude_dirs=exclude_dirs,
                                exclude_patterns=exclude_patterns,
                                add = add
                                )
    allowed = allowed or make_allowed_predicate(cfg)
    directories = make_list(directory)
    roots = [r for r in directories if r]

    # your existing helpers (get_dirs, get_globs, etc.) stay the same
    original_dirs = get_allowed_dirs(roots, allowed=allowed)
    original_globs = get_globs(original_dirs)
    files = get_allowed_files(original_globs, allowed=allowed)

    for d in get_filtered_dirs(original_dirs, allowed=allowed):
        files += get_filtered_files(d, allowed=allowed, files=files)

    # de-dupe while preserving order
    seen, out = set(), []
    for f in files:
        if f not in seen:
            seen.add(f)
            out.append(f)
    return out


def _fast_walk(
    root: Path,
    exts: Iterable[str],
    skip_dirs: Iterable[str] = (),
    skip_patterns: Iterable[str] = (),
) -> List[Path]:
    exts = tuple(exts)
    skip_dirs = set(sd.lower() for sd in skip_dirs or ())
    skip_patterns = tuple(sp.lower() for sp in (skip_patterns or ()))

    out = []
    for p in root.rglob("*"):
        # skip directories by name hit
        if p.is_dir():
            name = p.name.lower()
            if name in skip_dirs:
                # rglob doesn't let us prune mid-iteration cleanly; we just won't collect under it
                continue
            # nothing to collect for dirs
            continue

        # file filters
        name = p.name.lower()
        if any(fnmatch.fnmatch(name, pat) for pat in skip_patterns):
            continue
        if p.suffix.lower() in exts:
            out.append(p)

    # de-dup and normalize
    return sorted({pp.resolve() for pp in out})


def enumerate_source_files(
    src_root: Path,
    cfg: Optional["ScanConfig"] = None,
    *,
    exts: Optional[Iterable[str]] = None,
    fast_skip_dirs: Optional[Iterable[str]] = None,
    fast_skip_patterns: Optional[Iterable[str]] = None,
) -> List[Path]:
    """
    Unified enumerator:
      - If `cfg` is provided: use collect_filepaths(...) with full rules.
      - Else: fast walk using rglob over `exts` (defaults to EXTS) with optional light excludes.
    """
    src_root = Path(src_root)

    if cfg is not None:
        files = collect_filepaths([str(src_root)], cfg=cfg)
        return sorted({Path(f).resolve() for f in files})

    # Fast mode
    return _fast_walk(
        src_root,
        exts or EXTS,
        skip_dirs=fast_skip_dirs or (),
        skip_patterns=fast_skip_patterns or (),
    )
